@@ -31,6 +31,9 @@ static void     html_box_block_text_accessible_real_initialize (AtkObject *objec
 static gint     html_box_block_text_accessible_get_n_children  (AtkObject *obj);
 static AtkObject*   html_box_block_text_accessible_ref_child   (AtkObject *obj,
                                                                 gint      i);
+static AtkRelationSet* html_box_block_text_accessible_ref_relation_set
+                                                               (AtkObject *obj)
+;
 
 static void     html_box_block_text_accessible_text_interface_init (AtkTextIface	*iface);
 static gchar*   html_box_block_text_accessible_get_text            (AtkText             *text,
@@ -101,11 +104,14 @@ static gchar*   get_text_near_offset                               (AtkText
                                                                     gint              *start_offset,
                                                                     gint              *end_offset);
 
+extern HtmlBoxText* _html_view_get_cursor_box_text (HtmlView *view, gint *offset);
+
 static gpointer parent_class = NULL;
 
 struct _HtmlBoxBlockTextAccessiblePrivate
 {
 	GailTextUtil *textutil;
+	gint          caret_offset;
 };
 
 GType
@@ -165,6 +171,7 @@ html_box_block_text_accessible_class_init (HtmlBoxBlockAccessibleClass *klass)
 	class->initialize = html_box_block_text_accessible_real_initialize;
 	class->get_n_children = html_box_block_text_accessible_get_n_children;
 	class->ref_child = html_box_block_text_accessible_ref_child;
+	class->ref_relation_set = html_box_block_text_accessible_ref_relation_set;
 }
 
 static void
@@ -187,8 +194,6 @@ append_text (HtmlBox *root,
 
 	if (!root)
 		return;
-	if (HTML_IS_BOX_TABLE (root))
-		return;
 
 	if (HTML_IS_BOX_TEXT (root)) {
 		text = html_box_text_get_text (HTML_BOX_TEXT (root), &len);
@@ -233,73 +238,172 @@ html_box_block_text_accessible_real_initialize (AtkObject *object,
 static gint
 html_box_block_text_accessible_get_n_children (AtkObject *obj)
 {
-        AtkGObjectAccessible *atk_gobject;
-        HtmlBox *box;
-        gint n_children = 0;
-        GObject *g_obj;
-
         g_return_val_if_fail (HTML_IS_BOX_ACCESSIBLE (obj), 0);
-        atk_gobject = ATK_GOBJECT_ACCESSIBLE (obj);
-        g_obj = atk_gobject_accessible_get_object (atk_gobject);
-        if (g_obj == NULL)
-                return 0;
-
-        g_return_val_if_fail (HTML_IS_BOX (g_obj), 0);
-        box = HTML_BOX (g_obj);
-
-        if (box) {
-                HtmlBox *child;
-
-                child = box->children;
-
-                while (child) {
-                        if (HTML_IS_BOX_TABLE (child))
-				n_children++;
-                        child = child->next;
-                }
-        }
-        return n_children;
+	return 0;
 }
 
 static AtkObject*
 html_box_block_text_accessible_ref_child (AtkObject *obj,
                                           gint      i)
 {
-        AtkGObjectAccessible *atk_gobject;
-        GObject *g_obj;
-        HtmlBox *box;
-        AtkObject *atk_child = NULL;
-        gint n_children = 0;
-
         g_return_val_if_fail (HTML_IS_BOX_ACCESSIBLE (obj), NULL);
-        atk_gobject = ATK_GOBJECT_ACCESSIBLE (obj);
-        g_obj = atk_gobject_accessible_get_object (atk_gobject);
-        if (g_obj == NULL)
-                return NULL;
+	return NULL;
+}
 
-        g_return_val_if_fail (HTML_IS_BOX (g_obj), NULL);
-        box = HTML_BOX (g_obj);
+static AtkObject *
+ref_last_child (AtkObject *obj)
+{
+	AtkObject *child;
+	gint n_children;
 
-        if (box) {
-                HtmlBox *child;
+	n_children = atk_object_get_n_accessible_children (obj);
+	if (n_children > 0) {
+		child = atk_object_ref_accessible_child (obj, n_children - 1);
 
-                child = box->children;
+		while (child) {
+			n_children = atk_object_get_n_accessible_children (child);
+			if (n_children > 0) {
+				g_object_unref (child);
+				child = atk_object_ref_accessible_child (child, n_children -1);
+			} else {
+				break;
+			}
+		}
+		return child;
+	} else {
+		return NULL;
+	}
+}
 
-                while (child) {
-                        if (HTML_IS_BOX_TABLE (child)) {
-                        	if (n_children == i) {
-                               		atk_child = atk_gobject_accessible_for_object (G_OBJECT (child));
-					g_assert (HTML_IS_BOX_ACCESSIBLE (atk_child));
-					HTML_BOX_ACCESSIBLE (atk_child)->index = i;
-					g_object_ref (atk_child);
-					break;
+static AtkObject*
+ref_previous_object (AtkObject *obj)
+{
+	AtkObject *parent;
+	AtkObject *prev;
+	AtkObject *tmp;
+	gint index;
+	gint n_children;
+
+	index = atk_object_get_index_in_parent (obj);
+	parent= atk_object_get_parent (obj);
+	if (!HTML_IS_BOX_ACCESSIBLE (parent))
+		return NULL;
+
+	if (index > 0) {
+		n_children = atk_object_get_n_accessible_children (obj);
+		prev = atk_object_ref_accessible_child (parent, index - 1);
+		tmp = ref_last_child (prev);
+		if (tmp) {
+			g_object_unref (prev);
+			prev = tmp;
+		}
+	} else {
+		prev = parent;
+		while (prev) {
+			index = atk_object_get_index_in_parent (prev);
+			parent= atk_object_get_parent (prev);
+			if (!HTML_IS_BOX_ACCESSIBLE (parent))
+				return NULL;
+
+			if (index > 0) {
+				n_children = atk_object_get_n_accessible_children (obj);
+				prev = atk_object_ref_accessible_child (parent, index - 1);
+				tmp = ref_last_child (prev);
+				if (tmp) {
+					g_object_unref (prev);
+					prev = tmp;
 				}
-                        	n_children++;
-			}	
-                        child = child->next;
-                }
-        }
-        return atk_child;
+				break;
+			} else {
+				prev = parent;
+			}
+		}
+	}
+	return prev;
+}
+
+static AtkObject*
+ref_next_object (AtkObject *obj)
+{
+	AtkObject *parent;
+	AtkObject *next;
+	gint index;
+	gint n_children;
+
+	n_children = atk_object_get_n_accessible_children (obj);
+	if (n_children) {
+		return atk_object_ref_accessible_child (obj, 0);
+	}
+	parent = atk_object_get_parent (obj);
+	if (!HTML_IS_BOX_ACCESSIBLE (parent))
+	        return NULL;
+
+	index = atk_object_get_index_in_parent (obj);
+	n_children = atk_object_get_n_accessible_children (parent);
+	if (index < n_children - 1) {
+		return atk_object_ref_accessible_child (parent, index + 1);
+	} else {
+		next = parent;
+		while (next) {
+			parent = atk_object_get_parent (next);
+			if (!HTML_IS_BOX_ACCESSIBLE (parent))
+				return NULL;
+
+			index = atk_object_get_index_in_parent (next);
+			n_children = atk_object_get_n_accessible_children (parent);
+			if (index < n_children - 1) {
+				return atk_object_ref_accessible_child (parent, index + 1);
+			} else {
+				next = parent;
+			}
+		}
+		return NULL;
+	}
+}
+
+static AtkRelationSet*
+html_box_block_text_accessible_ref_relation_set (AtkObject *obj)
+{
+	AtkRelationSet *relation_set;
+	AtkObject *atk_obj;
+	AtkRelation *relation;
+	AtkObject *accessible_array[1];
+
+	relation_set = ATK_OBJECT_CLASS (parent_class)->ref_relation_set (obj);
+	if (!atk_relation_set_contains (relation_set, ATK_RELATION_FLOWS_TO)) {
+		atk_obj = ref_next_object (obj);
+		while (atk_obj) {
+			if (ATK_IS_TEXT (atk_obj))
+				break;
+			g_object_unref (atk_obj);
+			atk_obj = ref_next_object (atk_obj);
+		}
+		if (atk_obj) {
+			g_object_unref (atk_obj);
+			accessible_array [0] = atk_obj;
+			relation = atk_relation_new (accessible_array, 1, ATK_RELATION_FLOWS_TO);
+			atk_relation_set_add (relation_set, relation);
+			g_object_unref (relation);
+		}
+	}
+	if (!atk_relation_set_contains (relation_set, ATK_RELATION_FLOWS_FROM)) {
+		atk_obj = ref_previous_object (obj);
+		while (atk_obj) {
+			if (ATK_IS_TEXT (atk_obj))
+				break;
+			g_object_unref (atk_obj);
+			atk_obj = ref_previous_object (atk_obj);
+		}
+		if (atk_obj) {
+			g_object_unref (atk_obj);
+			accessible_array [0] = atk_obj;
+			relation = atk_relation_new (accessible_array, 1, ATK_RELATION_FLOWS_FROM);
+			atk_relation_set_add (relation_set, relation);
+			g_object_unref (relation);
+		}
+
+	}
+	return relation_set;
 }
 
 static void
@@ -307,7 +411,6 @@ html_box_block_text_accessible_text_interface_init (AtkTextIface *iface)
 {
  	g_return_if_fail (iface != NULL);
 
-	iface->get_text = html_box_block_text_accessible_get_text;
 	iface->get_text = html_box_block_text_accessible_get_text;
 	iface->get_text_after_offset = html_box_block_text_accessible_get_text_after_offset;
 	iface->get_text_at_offset = html_box_block_text_accessible_get_text_at_offset;
@@ -415,28 +518,78 @@ html_box_block_text_accessible_get_character_count (AtkText *text)
 	HtmlBoxBlockTextAccessible *block;
 	GtkTextBuffer *buffer;
 
-	g_return_val_if_fail (HTML_BOX_BLOCK_TEXT_ACCESSIBLE (text), NULL);
+	g_return_val_if_fail (HTML_BOX_BLOCK_TEXT_ACCESSIBLE (text), 0);
 	block = HTML_BOX_BLOCK_TEXT_ACCESSIBLE (text);
-	g_return_val_if_fail (block->priv->textutil, NULL);
+	g_return_val_if_fail (block->priv->textutil, 0);
 	buffer = block->priv->textutil->buffer;
 	return gtk_text_buffer_get_char_count (buffer);
 }
 
+static gboolean
+find_offset (HtmlBox *box, HtmlBoxText *box_text, gint *offset)
+{
+	HtmlBox *child;
+	HtmlBoxText *text;
+	gchar *char_text;
+	gint len;
+	gboolean ret;
+
+	if (HTML_IS_BOX_TEXT (box)) {
+		text = HTML_BOX_TEXT (box);
+		if (box_text == text)
+			return TRUE;
+	
+		char_text = html_box_text_get_text (text, &len);
+		len = g_utf8_strlen (char_text, len);
+		*offset += len;
+	}
+	child = box->children;
+	while (child) {
+		ret = find_offset (child, box_text, offset);
+		if (ret)
+			return ret;
+		child = child->next;
+	}		
+	return FALSE;
+}
+
 static gint
 html_box_block_text_accessible_get_caret_offset (AtkText *text)
 {
 	HtmlBoxBlockTextAccessible *block;
-	GtkTextBuffer *buffer;
-	GtkTextMark *cursor_mark;
-	GtkTextIter cursor_itr;
+        HtmlBox *box;
+        HtmlBoxText *cursor_box_text;
+        HtmlBox *cursor_box;
+        GtkWidget *widget;
+        HtmlView *view;
+        GObject *g_obj;
+        gint offset;
 
-	g_return_val_if_fail (HTML_BOX_BLOCK_TEXT_ACCESSIBLE (text), NULL);
+	g_return_val_if_fail (HTML_BOX_BLOCK_TEXT_ACCESSIBLE (text), 0);
 	block = HTML_BOX_BLOCK_TEXT_ACCESSIBLE (text);
-	g_return_val_if_fail (block->priv->textutil, NULL);
-	buffer = block->priv->textutil->buffer;
-	cursor_mark = gtk_text_buffer_get_insert (buffer);
-	gtk_text_buffer_get_iter_at_mark (buffer, &cursor_itr, cursor_mark);
-	return gtk_text_iter_get_offset (&cursor_itr);
+        g_obj = atk_gobject_accessible_get_object (ATK_GOBJECT_ACCESSIBLE (text
+));
+        if (g_obj == NULL)
+                return 0;
+
+        box = HTML_BOX (g_obj);
+        widget = html_box_accessible_get_view_widget (box);
+        view = HTML_VIEW (widget);
+        cursor_box_text = _html_view_get_cursor_box_text (view, &offset);
+	if (HTML_IS_BOX (cursor_box_text)) {
+		cursor_box = HTML_BOX (cursor_box_text);
+		while (cursor_box && !HTML_IS_BOX_BLOCK (cursor_box)) {
+			cursor_box = cursor_box->parent;
+		}
+		if (cursor_box == box) {
+			if (find_offset (box, cursor_box_text, &offset)) {
+                		block->priv->caret_offset = offset;
+			} else {
+				g_assert_not_reached ();
+			}
+		}
+	}
+        return block->priv->caret_offset;
 }
 
 static gboolean
@@ -447,13 +600,11 @@ html_box_block_text_accessible_set_caret_offset (AtkText *text,
 	GtkTextBuffer *buffer;
 	GtkTextIter pos_itr;
 
-	g_return_val_if_fail (HTML_BOX_BLOCK_TEXT_ACCESSIBLE (text), NULL);
+	g_return_val_if_fail (HTML_BOX_BLOCK_TEXT_ACCESSIBLE (text), FALSE);
 	block = HTML_BOX_BLOCK_TEXT_ACCESSIBLE (text);
-	g_return_val_if_fail (block->priv->textutil, NULL);
-	buffer = block->priv->textutil->buffer;
-	gtk_text_buffer_get_iter_at_offset (buffer,  &pos_itr, offset);
-	gtk_text_buffer_move_mark_by_name (buffer, "insert", &pos_itr);
-	return TRUE;
+	g_return_val_if_fail (block->priv->textutil, FALSE);
+	g_warning ("set_caret_offset not implemented");
+	return FALSE;
 }
 
 static gboolean
@@ -1044,7 +1195,7 @@ html_box_block_text_accessible_add_selection (AtkText *text,
 	start_text = find_box_text_for_offset (box, &start_offset);
 	end_text = find_box_text_for_offset (box, &end_offset);
 	if (!start_text)
-		return NULL;
+		return FALSE;
 
 	box_text = start_text;
 	while (box_text) {
